Fun things to do with the rendering test

A quick scene description language primer
=========================================

Before you can define any objects, you need to start a room.  Only the
alphabetically-first room will actually be rendered for now, because I
haven't gotten around to implementing portals (not that there's any
way to make proper doorways yet anyway).  As you can probably see at
the top of any of the scene files, it's simply just

	room testroom

(though you can call it whatever you want).

You can also define a camera with the 'camera' command.  It takes the
following parameters:

	pos - set the position
	up - set the direction of, well, up (note that this value will only
		be good for the first frame)
	lookat - set the position it's looking at
	fov - set the vertical field of view (in degrees)
	hith - set the near clipping plane
	yon - set the far clipping plane

A note on datatypes: the command parser has a few different datatypes
that it expects.  They are:

	vector (x,y,z)
	color (r,g,b,a)
	float - i.e. a single floating-point number
	bool (t/1/y for true, anything else for false)
	string (no whitespace accepted, not even with quotes)
	enumeration (specific to what's being enumerated)

Note that for the n-tuples (vector and color), all parameters after
the first one are optional, and will default to 0 (except alpha in
color, which defaults to 1).  So if you want to use the color black,
you can just say "0", and if you want the color green you can just say
"0,1", and so on.

So anyway.  Now we can create objects.  There are two ways to do this.
The first form is good if you want to set a whole bunch of things at
once:

	object [objectname]
	set [parameter] [value]
	set [parameter] [value]
	set [parameter] [value]
	...

and the second form is good if you want to just set one or two
parameters, such as if you're doing something interactively:

	set [objectname] [parameter] [value]

objectname is hierarchal, with / as the directory separator (there's
no need for it to be that way, that's just how I've coded it into the
command processor; the real scene format will probably be an XML of
some sort, though the attributes will probably be the same as what's
described so this document will still be somewhat useful then).

So, if you have objects foo, foo/bar, and foo/baz, foo/bar and foo/baz
will be moved relative to foo.  Rotating foo will cause foo/baz and
foo/bar to rotate around foo's center.  Certain properties (such as
'active' and 'billboard') will be inherited (implicit to their
nature), though most won't.

Here's the properties you can set (note that some of them really are
commands, hidden as properties; the command parser really is a
horrible hack):

	mesh [filename] - the base mesh for the node (currently this is
		just a relative or absolute path to the file, but later
		it'll be a URI)
	pos [vector] - the position of the object
	up [vector] - the direction the object will try to point its Y axis
	point [vector] - the direction the object will point its Z axis
	visible [bool] - whether this object's shaders should be executed
	shadow [bool] - whether this object should cast a shadow
	billboard [bool] - whether to keep the object pointing at the camera

Note that all non-distorting object transformations can be specified
by a pair of up and point vectors.  You can derive these from any
transformation by creating a rotation matrix, and then using the Y
axis row for "up" and the Z axis row for "point."
	
And now the commands which are executed through the 'set' command:

	shade [passname] - add the current shader to the specified pass
	clear [passname] - remove all shaders to the pass
	pop [passname] - remove the last-added shader to the pass

Wait, so we know how to add shaders to the shader scripts, but we
don't know how to make the shaders...  First, you should probably read
the slides to the (very basic) presentation I gave on this aspect of
my renderer, at
http://www.cs.nmsu.edu/~joshagam/Solace/papers/gras-02/ - it explains
what the three passes are, but to summarize, they are:

	base - opaque, lighting-independent stuff
	light - opaque, lighting-dependent stuff
	blend - transparent stuff which is lighting-dependent but can't get
		shadows cast on it
	shadow - stuff to be run after a shadow is cast

Anyway.  Defining a shader is where the fun comes in.  Here's the
commands for that:

	shader new - Reset the current shader to the default (shiny
		opaque white plastic)
	shader set [property] [value] - Set some aspect of the shader

So, here are the available properties: (enumerations will be explained
below)

	mesh [filename] - the mesh used to render this shader
	texture [filename] - the texturemap to multiply this shader by
	ambient [color] - set the ambient lighting color
	diffuse [color] - set the diffuse lighting color
	color [color] - synonym for setting both ambient and diffuse to
		the same value
	specular [color] - set the color of specular hilights
	shininess [float] - how tight the hilights should be
	emission [color] - how much it should glow
	blend [blendfunc] - what pixel operation to perform
	rmode [renderfunc] - what mode to render the mesh in
	amode [alphafunc] - what alpha test to perform
	athresh [float] - the threshold for the alpha test
	thick [float] - the thickness of lines for line-drawing rmodes
	smooth [float] - whether to smooth-shade the triangles
	lit [bool] - whether to apply lighting calculations to this shader
		(if not, it just uses the diffuse color for everything)
	faces [faceset] - which set of faces to draw
	depth [bool] - whether to insert this into the depth buffer

Okay, so the enumerations:

	blendfunc:
		opaque - if this is the first pass to be drawn for this
			object, do 'replace'; otherwise, do 'add'
		replace - just set the color, ignoring whatever is under it
		add - add the color to whatever's under it
		blend - traditional alpha blend
		mul - multiply the color by whatever's under it
		haze - add the color multiplied by the alpha
		filter - mul + haze
		subtract - subtract the color from whatever's under it

	renderfunc:
		solid - render the mesh as solid triangles
		wire - render the mesh as a hollow wireframe
		outline - render the outline of the mesh
		shadow - render the shadow volume as projected from the
			current light
		wireshadow - render the wireframe of the shadow volume

	alphafunc:
		1 - always draw the pixel
		< - draw the pixel if its alpha value is less than the
			threshold
		> - draw the pixel if its alpha value is greater than
			the threshold
		0 - never draw the pixel (these passes are typically
			just thrown out by the backend since they
			have no visible effect)

	faces:
		front - Draw front-facing faces
		back - Draw back-facing faces
		all - Draw all faces

A note on blend functions, and what's (normally) appropriate for each
pass (unless you know what you're doing or are feeling experimental):

	base: start with opaque or replace, anything goes after that
	light: start with opaque, use only opaque or add after that
	blend: anything goes

Anyway, now there's one last thing you need to know before you can
start making scenes: lights!  There are four commands for dealing
with lights: (in all commands, the room name can be left out, and will
just default to whichever was the last one specified with the 'room'
command)

	light ambient [room] [color] - the amount of light cast on all
		objects in the room (which defaults to the current
		room if it's left out) regardless of what lights are
		shining

	light set [room] [name] [property] [value] - set the given
		property of the specified light to the given value
		(creating a new light if one by that name doesn't exist)

	light delete [room] [name] - delete the given light

Here are the properties which can be set on a light:

	active [bool] - whether this light is shining or not
	pos [vec] - the position it's shining from
	color [color] - yeah
	falloff [vec] - the parameters for the lighting falloff
		equation, given as x + y*d + z*d^2 (where d is the distance
		from the light, and x, y and z are the values in this
		vector). Real lights behave with a falloff of 0,0,1
		but that usually doesn't look right in computer
		graphics
	ambient [color] - the amount of ambient light it adds to any objects
		it shines on (tricky; more on this later, but it
		should normally be set to 0)
	global [color] - the amount of light this contributes to the
		global ambient light when it's active
	angle [float] - the angle (in degrees) of the cone defining
		the spotlight from its axis; 90 is a flat plane, and
		180 is a special case for "shines in all directions
		equally"
	focus [float] - how tightly to focus the spotlight beam
	pointat [vec] - the position in space to point at
	shade [cmd] - similar to the "shade" pseudoproperty for objects;
		[cmd] is one of add, clear, or pop.  The shader script
		associated with a light is run on any object which casts
		a shadow from that light at the end of that light's
		render pass

Okay, that's all of the commands you need for creating a scene.  Now
there's a few more commands you should know, which only make sense
when used interactively:

	delete [room] [object] - delete an object from a room (in this
		case, 'room' is not optional, sorry; otherwise there'd
		be no way to delete a room with this crappy parser)

	reload [filename] - force the renderer to reload whatever
		resource is at that filename (like, if you modify a
		mesh or texture and want to see it without restarting,
		just reload its filename).  You can also use this to
		fake the progressive streaming of a mesh or texture to
		see what it'll look like as it downloads (see later in
		this file)

	emit [filename] - save the current scenegraph to a file for
		later (though the logic for this is a bit stupid,
		especially with shaders, so don't be surprised if it
		stores a lot of unnecessary stuff)

	r_set [key] [value] - Show and/or set the settings available
		for the renderer.  If value is omitted, it displays
		the setting, and if key is omitted, it displays all
		settings (including textual descriptions)

	d_set [key] [value] - The same, but for the driver

Something to note... you might notice that the node itself has a mesh,
and the individual shaders can have a mesh as well.  This is because
the node's mesh is used for visibility determination and shadow
casting, and the shaders' meshes are what's used for rendering.  If
the shader doesn't have a mesh of its own it just uses the node's
mesh, but keep the following in mind:

	If a node doesn't have a mesh, it will never be visible or
	cast a shadow

	A shader's visibility will be determined based on the node's
	mesh, not the shader (see the scene test2 for an example of
	this - the shimmery ball of haze is only visible when the
	"nucleus" of the Solace logo is)

	The shadow which is cast will always be the node's, never the
	shader's (so a sphere node with a torus shader will look like
	a torus but will cast a sphere's shadow)

Most of the time you don't want to specify a mesh for a shader, unless
you're doing something tricky (for example, for a light's corona, it's
nifty to give the node a small-ish sphere for its mesh but then use
the standard billboard mesh for the shader; the objects light3/flare
and light4/flare in stress2 use this trick).  Typically, if you have
an object with different meshes for some parts, you'll want to use a
mesh whose bounding box contains all of the part meshes.

Also, note that if you associate a shader with a light, the shader
will always use the object's base mesh (unless the shader itself has a
mesh associated with it, in which case the shader's mesh will be used
on all objects).

Also note that the occlusion buffer basically runs the entire shader
script (taking many shortcuts wherever possible), so it's the parts
which are actually drawn which affect occlusion, not the node's base
mesh.

To summarize: only explicitly specify a mesh for a shader if you know
what you're doing. :)

scenes/cage:
============

From the console, turn on and off the lights; for example,

	light set light3 active 0
	light set light1 active 1

and notice how it affects the global illumination.  This is actually
manually written into the scene.  For example, you can make it so that
light 3 being active floods the room with red light by doing

	light set light3 global 1,0,0

Also, colors don't have to be positive; for example, if you want a
light to "suck up" all non-shadowing light, try something like

	light set light2 global -1000,-1000,-1000

or

	light set light2 color -1,-1,-1

scenes/test6:
=============

Turn down the mesh detail and watch the spikey ball pulsate due to
sample aliasing.  This effect could actually be taken advantage of, if
desired.

scenes/stress*
==============

Try turning on and off wireframe (with 'w') and occlusion culling
(with 'o') to see the effect it has on invisible geometry and the
framerate.

In stress1, try moving the "jiggler" object somewhere outside the room
and see it get culled out nicely (using wireframe mode)... for
example,

	set jiggler pos 30,30,30

Or just put it barely outside the room with

	set jiggler pos 0,0,20

and watch it slowly invade from behind. :)

Also in stress1, notice how turning on shadows causes the green light
to be blocked by the one-way partition, but not the white light.  This
is because shadows are computed based on what's visible to the object
receiving the shadow to make one-way objects at least a little less
unintuitive.

scenes/test8:
=============

The spheres' shader does an evil trick.  Its first lit layer is
transparent.  The renderer expects the first lit layer to be opaque.
This causes some really weird effects.  For example, with every
shadowing light, the lit parts get more and more opaque.
Additionally, it really messes with the occlusion buffer; any objects
which are totally behind something like that will tend to flicker in
and out of visibility.  (For example, look at the row of spheres from
one end, or get really close to a sphere and look at the cube through
it.)

Also, such objects are only sorted correctly if they were visible in
the previous frame and haven't moved or changed since then.  (It's a
natural side effect of how the occlusion buffer is implemented.
Technically I could have implemented it such that their draw order
would be the same no matter what, but it uses the came code as the
multipass rendering stuff, which sorts it to try to deal with
properly-transparent objects anyway, and the overhead isn't enough to
worry about.)  And if the occlusion buffer is turned off, they're just
left unsorted, which means they're ordered alphabetically (since
that's the order they're traversed in the scenegraph, since the name
is the primary key for the tree).

Also, if the client were set to not clear the color buffer every
frame, it would lead to many odd side-effects, such as having motion
blur occurring inside certain objects (but only if they're drawn
before everything else).

So a lot of different aspects make transparent base/lit passes
somewhat unpredictable, though in cool ways. :)

general
=======

Although the renderer itself doesn't (currently) have anything for
attaching lights to objects, making objects move, and so on, there are
a few special-cased behaviors in the testRender client itself.

An object named the same as a light will always be oriented around the
light.

A light whose name starts with : will fly around (as in the scene
"cage-animated"); note that the objects also have to be named with the
: prefix.

A toplevel object named "jiggler" will have its contents jiggle around
randomly (as in the scene "stress").

cartoon shading (and other lighting tricks)
===========================================

There are two good (but somewhat slow) ways to perform cartoon
lighting tricks (that is, the shading of cartoony objects).  Both of
them abuse the fact that non-global ambient light still has shadows
cast on it (this was a conscious design decision, specifically to make
these tricks possible).

The first way applies the cartoon lighting to a single object.  Simply
create an object which has a lighting pass with emission, like this:

	object cartoon-shaded-sphere
	set mesh meshes/sphere.nmm
	shader new
	shader set color 0,0,0
	shader set emission 1,0,0
	shader set specular 0,0,0
	set shade light

This is the technique used by most of the cartoon-shaded objects right
now.  However, you can also apply cartoon shading to an entire scene
(or, more precisely, to an entire light):

	light set cartoon-light color 0,0,0
	light set cartoon-light ambient 1,1,1

This will cause all objects lit by that light to glow with their
ambient color instead of being shaded normally.  However, mixing
normal and cartoony lights in a single room isn't recommended, nor is
having cartoonly spotlights.  For a taste of what happens when you do
this, try loading up the scene "cage-animated" and issue these
commands:

	light set :light1 color 0
	light set :light1 ambient 0,0.5,1

With this sort of trick you can make objects which glow a different
color than they're normally lit.  When you set the color of a shader,
it's really a synonym for "set both diffuse and ambient."  But if you
want an object which will be lit with blue but will glow with red
(i.e. with the global ambient light, or when hit by a light which
provides its own ambient), do something like:

	object weird-glowing
	set mesh meshes/sphere.nmm
	shader new
	shader set diffuse 0,0,1
	shader set ambient 1,0,0
	set shade light

You can also do some pretty complex tricks with negative colors; look
at the shader for the skydome and the pointy object in the scene
"test3" for an example:

	shader new
	shader set emission 1,0,1
	shader set color -1,1,0
	shader set specular -1,1,-1
	set shade light

So, it glows magenta, but direct light makes it add green and remove
red (turning it cyan), but the specular hilight removes red and blue,
turning it green.  Note that colors which are negative,
super-saturated (greater than 1), or otherwise weird usually behaves
differently with a shadowing light than with a non-shadowing light
(since non-shadowing lights are combined into as few passes as
possible).

Finally, there's a third (but not highly-recommended) way to do
per-object cartoon shading, where you create a shader like this (for
example):

	shader new
	shader set diffuse 0
	shader set ambient 1,0,0
	shader set specular 1,0,0
	shader set shininess 0

This has the advantage of not requiring shadows to be on to have the
directional shading, but has the disadvantage of not looking very good
(at least, not in my opinion; I prefer a hard edge, which you can sort
of get by doing 'shader set smooth 0' but it still looks kinda bad)).
But since it doesn't rely on shadows, you can use it in the blend
pass, if you want, for example, a toon-shaded halo around something,
or want to get the toon-shaded look without requiring shadows.  You
can see what this technique looks like in the scene "alternate-toon,"
in any case (the specular method on the left, the emissive method on
the right).

Oh, and of course, that leaves one last thing for cartoon shading: the
outline!  Having this as your very last shader will do what you want:

	shader new
	shader set color 0
	shader set lit 0
	shader set rmode outline
	shader set thick 5

There's more complex layered effects you can do by applying a texture
and lighting and so on to the line, but if you just want a thick black
outline, you can't do wrong with that one. :)

Shadow shaders
==============

Shadow shaders are a rather unusual thing you can do for some pretty
interesting effects.  "Shadow shaders" as a whole refers to either
per-object shaders (where the render function is given as shadow or
wireshadow) or per-light shaders.  They share some things in common,
but they also have some pretty bizarre behavior on their own.

For example, per-object shadow shaders only work if there's exactly
one light shining at the time of render.  This means that either the
room has to have exactly one light, or the scene has to be rendered
with shadows enabled.  (When shadows are enabled, it really rerenders
the scene a bunch of times for each light.  Part of the magic of
Solace is the massive shortcuts it takes to make this not suck. :)

Per-light shadow shaders apply to all objects which cast a shadow from
the light with the shader, and always use the shader's mesh if it's
specified, or the object's base mesh if not.

With that in mind, let's take the earlier example of a "cartoon
light," and extend it to add an outline around all shadow-casting
objects (which will hopefully be everything which would end up looking
toon-shaded anyway):

	light set toon color 0
	light set toon ambient 1,1,1
	shader new
	shader set thick 5
	shader set rmode outline
	shader set color 10,10,10
	shader set blend subtract
	light set toon shade add

So what this does is makes a light (called "toon") which casts white
ambient light on everything (just as before), only now it also casts a
shader on them as well.  Specifically, it's a shader which subtracts a
really bright white outline of the object from the image.  (This way
it'll fade out as it leaves the influence of the light.)

Another nifty thing you can do is to make the shadow volume visible in
a number of styles.  For that you'd make a shader like this (which can
be associated either with the light or an individual object):

	shader new
	shader set rmode shadow
	shader set color 1,1,1,0.1
	shader set blend blend
	shader set depth 0

This will basically draw a semi-faint white "shell" around the edge of
the shadow volume.

Note that setting depth to 0 is pretty important.  Otherwise, anything
which is drawn behind this shadow volume on subsequent passes won't
appear.  Unfortunately, setting depth to 0 means that other things
will mess up in odd ways (for example, objects drawn behind it from
another room will clobber it, and other transparent objects drawn in
the light pass will interact strangely with it), though for the most
part things will behave correctly.

Now, if you don't set depth to 0, you can get some pretty nifty
effects anyway.  However, they won't be stable, and they'll depend
quite a bit on what order things are drawn in.  In practice this isn't
a problem, but in theory there are cases where this will really mess
things up.

Also, note that shadows won't be cast on things which are drawn with
depth set to 0.

As far as per-object shadow shaders, any shader script which involves
an rmode of shadow or wireshadow should go into the "shadow" pass,
rather than light or blend.  So, for example, here's a simple object
which shows off its shadow volume:

	object foo
	set mesh meshes/sphere.nmm
	// The object itself as a bubble
	shader new
	shader set color 0.3,0.1,0.1
	shader set specular 1,1,1
	shader set shininess 20
	shader set blend add
	set shade blend
	// The object's shadow volume as a wireframe
	shader new
	shader set rmode wireshadow
	set shade shadow

Note that shadow-pass shaders don't necessarily have to involve a
[wire]shadow rmode, but if the rmode is [wire]shadow, it should
probably go into the shadow pass.  (You can get some pretty
interesting effects by putting them into the light pass, however!)

Also, if an object has its own shadow shader, it will be run even if
the object doesn't normally cast a shadow.  This can be pretty useful
for doing certain effects; for example, this object casts an odd sort
of shadow where the shadow volume is magenta, but objects inside the
shadow volume are lit with blue.


	object weird-shadow
	set shadow 0
	set mesh meshes/sphere-2.0.nmm
	shader new
	shader set color 1,1,1,0.4
	shader set rmode outline
	shader set thick 3
	set shade light
	shader new
	shader set color 0.1,0,0,0.1
	shader set lit 0
	shader set depth 0
	shader set rmode shadow
	shader set faces back
	shader set blend add
	set shade shadow
	shader set color 0,0,0.1,0.1
	shader set faces front
	set shade shadow

It works by emulating the way that shadow volumes are rendered to
begin with.  By playing with the blend functions, you can get some
pretty interesting effects, as well; for example, replace both shaders
in the shadow pass with this for an interesting x-ray blacklight
effect:

	shader new
	shader set depth 0
	shader set color 1,0.5,1,0.3
	shader set blend filter
	shader set rmode shadow
	shader set faces front
	set shade shadow
	shader set color 0,0,0
	shader set blend mul
	shader set faces back
	set shade shadow

Note that similar techniques can be used for a similar effect on
ordinary shaders as well - they don't have to work on an rmode of
shadow.  You could just as easily have the object itself be a sphere
of x-ray vision, performed in the blend pass. :)

Another note: If there's both a per-light and a per-object shadow
shader, the per-light one runs first.

Progressive downloading
=======================

Although progressive downloading (or any sort of downloading, for that
matter) of resources isn't yet implemented, you can still see what it
looks like.  Here's an example of how to do it interactively... first,
run testRender in one window, and a normal shell in another window.
Then, take the file you want to "stream" and (from the shell)
partially copy it with, say,

	dd if=meshes/sphere.nmm of=/tmp/asdf.nmm bs=1024 count=1

then in testRender, create an object which references that partial
file, for example

	object streamy
	set mesh /tmp/asdf.nmm
	shader new
	set shade base

then you can see what the object looks like after 1K has downloaded.
Then in the shell window, do the same command, but modify it slightly:

	dd if=meshes/sphere.nmm of=/tmp/asdf.nmm bs=1024 count=2

and then in testRender,

	reload /tmp/asdf.nmm

and the object should immediately update with the 2k "download."  Keep
on incrementing the value of count and then reloading the mesh.

This also serves as a cheap-ass lossy compression - to choose the
quality of an image, just decide how big you want the file to be, and
then truncate it to that point. :)

Motion Paths
============

Motion paths in Solace are currently very limited and excruciatingly
kludgy.  Basically, moving lights and the camera can have motions
associated with them now.  The way you do this is by setting the
'point' properties of special objects to various parameters.  So, for
example, if you have a moving light named ':light1', you can set its
parameters with something like this:

	set :light1/c point 0,0,0
	set :light1/p point 1,200,200
	set :light1/k point 0.02,0.05,0.05
	set :light1/o point 5,0,1.571
	set :light1/t point 1,0,0

The way it works is this... motions are (currently) just defined as a
simple sin function, where for any given component, the motion is
specified as:

	c + p*sin(k*n + o)

where n is the frame number, taking the parameter for the respective
axis for each output (so for example, for the x axis it's c.x, p.x,
k.x, and o.x).  Then 't' is the blending function for the 'pointat'
parameter; x is the weight for the center of the motion, y is the
weight for the forward direction of motion, and z is the weight for
the previous frame's pointat.

For the camera's motion path, the respective object is _cam; so, for
example, to make the camera move in a large lissajous pattern around
the point 10,20,30 and always face forward, you'd do something like:

	set _cam/c point 10,20,30
	set _cam/p point 200,200,200
	set _cam/k point 0.05,0.061,0.93
	set _cam/o point 1,2,3
	set _cam/t point 0,1,0

scenes/tunnel2 is a pretty good example of how to do this stuff.

This will, of course, be totally deprecated when there's an actual
physics engine.
